"""
LangXchange: A modular Framework for working with LLMs, vector databases, and diverse data sources.
"""

# Optional: Set a default version (used by setup.py or scripts)
__version__ = "0.4.4"

# Import common helper classes for direct access from package root
# from .chroma_helper import ChromaHelper
# from .faiss_helper import FAISSHelper
# from .mysql_helper import MySQLHelper
# from .mongo_helper import MongoHelper
# from .file_helper import FileHelper
# from .openai_helper import OpenAIHelper
# from .google_genai_helper import GoogleGenAIHelper
# from .llama_helper import LLaMAHelper
# # from llm/deepseek_helper import DeepSeekHelper
# # from .anthropic_helper import AnthropicHelper
# from .llama_helper import LLaMAHelper
# from .google_drive_helper import GoogleDriveHelper
# from .google_cs_helper import GoogleCloudStorageHelper
# from .drive_helper import DriveHelper
# from .opensearch_helper import OpenSearchHelper
# from .data_format_cleanup_helper import DataFormatCleanupHelper

# from .retrieverX import RetrieverX
# from .PrompterResponse import PrompterResponse
# from .documentloader import DocumentLoaderHelper


# from .core import LangXchangeAPI
# # from .vectordb import VectorDBHelper
# # from .dataconnectors import DataConnector
# # from .datafetcher import DataFetcher
# from .chroma_helper import ChromaHelper
# # from .pinecone_helper import PineconeHelper
# from .mysql_helper import MySQLHelper
# from .mongo_helper import MongoHelper
# from .file_helper import FileHelper
# from .faiss_helper import FAISSHelper
# # from .elasticsearch_helper import ElasticSearchHelper
# from .opensearch_helper import OpenSearchHelper
# from .openai_helper import OpenAIHelper
# from .google_genai_helper import GoogleGenAIHelper
# from .llama_helper import LLaMAHelper
# from .drive_helper import DriveHelper
# from .google_cs_helper import GoogleCloudStorageHelper
# # from llm/deepseek_helper import DeepSeekHelper
# # from .anthropic_helper import AnthropicHelper